part of masamune.ui;

TabContext<T> useTab<T>(List<T> source) {
  final context = useContext();
  final controller = useTabController(
    initialLength: source.length,
    keys: [source.length],
  );
  return TabContext<T>._(
    context: context,
    controller: controller,
    source: source,
  );
}

extension WidgetRefTabControllerExtensions on WidgetRef {
  TabContext<T> useTab<T>(List<T> source) {
    return containerBuilder<TabContext<T>, TabContext<T>>(
      key: "tab",
      containerBuilder: (container) {
        container?.dispose();
        final controller = TabController(
          vsync: _TickerProviderContainer(),
          length: source.length,
        );
        return TabContext<T>._(
          context: this as BuildContext,
          controller: controller,
          source: source,
        );
      },
      resultBuilder: (container) {
        return container;
      },
      test: (container) => source.length != container.length,
    );
  }

  TabController useTabController({
    required int initialLength,
    TickerProvider? vsync,
    int initialIndex = 0,
  }) {
    return containerBuilder<_TabControllerContainer, TabController>(
      key: "tabController",
      containerBuilder: (container) {
        container?.dispose();
        final controller = TabController(
          initialIndex: initialIndex,
          vsync: vsync ?? _TickerProviderContainer(),
          length: initialLength,
        );
        return _TabControllerContainer(
          controller: controller,
          length: initialLength,
        );
      },
      resultBuilder: (container) {
        return container.controller;
      },
      test: (container) => container.length != initialLength,
    );
  }

  TickerProvider useSingleTickerProvider() {
    return containerBuilder<_TickerProviderContainer, TickerProvider>(
      key: "singleTickerProvider",
      containerBuilder: (container) {
        return _TickerProviderContainer();
      },
      resultBuilder: (container) {
        return container;
      },
    );
  }
}

@immutable
class _TabControllerContainer extends VariableContainer {
  _TabControllerContainer({
    required this.controller,
    this.length,
  });

  final TabController controller;
  final int? length;

  @override
  void didUpdateContainer(_TabControllerContainer oldContainer) {
    super.didUpdateContainer(oldContainer);
    if (controller != oldContainer.controller) {
      oldContainer.dispose();
    }
  }

  @override
  void dispose() {
    super.dispose();
    controller.dispose();
  }
}

class _TickerProviderContainer implements TickerProvider {
  _TickerProviderContainer();
  Ticker? _ticker;

  @override
  Ticker createTicker(TickerCallback onTick) {
    assert(() {
      if (_ticker == null) {
        return true;
      }
      throw FlutterError(
          'You attempted to use a TickerProviderContainer multiple times.\n'
          'A SingleTickerProviderStateMixin can only be used as a TickerProvider once. '
          'If you need multiple Ticker, consider using useSingleTickerProvider multiple times '
          'to create as many Tickers as needed.');
    }(), '');
    return _ticker =
        Ticker(onTick, debugLabel: 'created by TickerProviderContainer');
  }
}

@immutable
class TabContext<T> extends VariableContainer {
  TabContext._({
    required this.controller,
    required this.source,
    required BuildContext context,
  })  : length = source.length,
        _context = context;

  final TabController controller;
  final BuildContext _context;
  final int length;
  final List<T> source;

  @override
  void didUpdateContainer(TabContext<T> oldContainer) {
    super.didUpdateContainer(oldContainer);
    if (controller != oldContainer.controller) {
      oldContainer.dispose();
    }
  }

  @override
  void dispose() {
    super.dispose();
    controller.dispose();
  }
}

class UITabBar<T> extends TabBar {
  /// Creates a material design tab bar.
  ///
  /// The [tabs] argument must not be null and its length must match the [controller]'s
  /// [TabController.length].
  ///
  /// If a [TabController] is not provided, then there must be a
  /// [DefaultTabController] ancestor.
  ///
  /// The [indicatorWeight] parameter defaults to 2, and must not be null.
  ///
  /// The [indicatorPadding] parameter defaults to [EdgeInsets.zero], and must not be null.
  ///
  /// If [indicator] is not null or provided from [TabBarTheme],
  /// then [indicatorWeight], [indicatorPadding], and [indicatorColor] are ignored.
  UITabBar(
    TabContext tab, {
    Key? key,
    Widget? Function(BuildContext context, T item)? builder,
    bool isScrollable = false,
    Color? indicatorColor,
    bool automaticIndicatorColorAdjustment = true,
    double indicatorWeight = 2.0,
    EdgeInsetsGeometry indicatorPadding = EdgeInsets.zero,
    Decoration? indicator,
    TabBarIndicatorSize? indicatorSize,
    Color? labelColor,
    TextStyle? labelStyle,
    EdgeInsetsGeometry labelPadding =
        const EdgeInsets.symmetric(horizontal: 16, vertical: 12),
    Color? unselectedLabelColor,
    TextStyle? unselectedLabelStyle,
    DragStartBehavior dragStartBehavior = DragStartBehavior.start,
    MaterialStateProperty<Color?>? overlayColor,
    MouseCursor? mouseCursor,
    bool? enableFeedback,
    void Function(T item)? onTap,
    ScrollPhysics? physics,
  }) : super(
          tabs: tab.source.mapAndRemoveEmpty((item) {
            if (builder != null) {
              return builder.call(tab._context, item);
            }
            return Text(item.toString());
          }),
          controller: tab.controller,
          key: key,
          isScrollable: isScrollable,
          indicatorColor: indicatorColor,
          automaticIndicatorColorAdjustment: automaticIndicatorColorAdjustment,
          indicatorWeight: indicatorWeight,
          indicatorPadding: indicatorPadding,
          indicator: indicator,
          indicatorSize: indicatorSize,
          labelColor: labelColor,
          labelStyle: labelStyle,
          labelPadding: labelPadding,
          unselectedLabelColor: unselectedLabelColor,
          unselectedLabelStyle: unselectedLabelStyle,
          dragStartBehavior: dragStartBehavior,
          overlayColor: overlayColor,
          mouseCursor: mouseCursor,
          enableFeedback: enableFeedback,
          onTap: onTap != null ? (i) => onTap.call(tab.source[i]) : null,
          physics: physics,
        );
}

class UITabView<T> extends TabBarView {
  UITabView(
    this.tab, {
    Key? key,
    required this.builder,
    ScrollPhysics? physics,
    DragStartBehavior dragStartBehavior = DragStartBehavior.start,
  }) : super(
          key: key,
          children: tab.source.mapAndRemoveEmpty(
              (item) => builder.call(tab._context, item, PageStorageKey(item))),
          controller: tab.controller,
          physics: physics,
          dragStartBehavior: dragStartBehavior,
        );

  final TabContext tab;

  /// One widget per tab.
  ///
  /// Its length must match the length of the [TabBar.tabs]
  /// list, as well as the [controller]'s [TabController.length].
  final Widget? Function(BuildContext context, T item, Key key) builder;
}
